import tensorflow as tf
from tensorflow.keras import layers

class CNN(tf.keras.Model):

    def __init__(self, vocab_size, emb_dim=128, nb_filters=50,
                ffn_units=512, nb_classes=2, dropout_rate=0.1,
                training=True, name="cnn"):

        super(CNN, self).__init__(name=name)

        self.embedding = layers.Embedding(vocab_size, emb_dim)

        self.bigram = layers.Conv1D(
            filters=nb_filters, kernel_size=2, padding='same', activation='relu')

        self.trigram = layers.Conv1D(
            filters=nb_filters, kernel_size=3, padding='same', activation='relu')

        self.fourgram = layers.Conv1D(
            filters=nb_filters, kernel_size=4, padding='same', activation='relu')

        self.pool = layers.GlobalMaxPool1D()

        self.dense_1 = layers.Dense(units=ffn_units, activation='relu')
        self.dropout = layers.Dropout(rate=dropout_rate)

        if nb_classes == 2:
            self.last_dense = layers.Dense(units = 1, activation = 'sigmoid')
        else:
            self.last_dense = layers.Dense(units = nb_classes, activation = 'softmax')

    def call(self, inputs, training):
        x = self.embedding(inputs)
        x_1 = self.bigram(x)
        x_1 = self.pool(x_1)
        x_2 = self.trigram(x)
        x_2 = self.pool(x_2)
        x_3 = self.fourgram(x)
        x_3 = self.pool(x_3)

        merged = tf.concat([x_1, x_2, x_3], axis = -1) # (batch_size, 3 * nb_filters)
        merged = self.dense_1(merged)
        merged = self.dropout(merged, training)
        output = self.last_dense(merged)

        return output




from datetime import datetime
import sys
import os 
from sys import argv

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import json
import pymongo
import pandas as pd

from database.db_mongo import MongoDatabase
from database.db_neo4j import Neo4jDatabase
import utils.utils as utils

class BI(object):

    def __init__(self, environment):
        self.environment = environment
        self.utils = utils
        self.process_bi()

    def process_bi(self):
        self.open_database_connections()

        self.products_df = self.utils.get_dataframe_from_mongo(self.mongo_conn)

        self.close_mongo_connection()

    def open_database_connections(self):
        self.mongo = MongoDatabase(self.environment)
        self.mongo.connect()
        self.mongo_conn = self.mongo

        self.neo4j = Neo4jDatabase(self.environment)
        self.neo4j_conn = self.neo4j.connect()

    def close_mongo_connection(self):
        self.mongo.close_connection()

BI(argv[1] if len(argv) > 1 else 'dev')


import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../'))

import tensorflow_datasets as tfds
import pickle
import numpy as np
import tarfile
import gdown

import utils.utils as utils
from nkia.ml.cnn_nlp_model import CNN

class classifyProducts(object):

    def __init__(self):
        self.utils = utils
        
    def load_model_to_memory(self, model_name):
        if 'category_model' not in os.listdir():
            self.download_model(
                'https://drive.google.com/u/1/uc?export=download&confirm=-Lqx&id=18aLg_tx-QRPrZrnFKZVVopC_2GTUIeti', 'category_model.tar.gz')

        if 'food_model' not in os.listdir():
            self.download_model(
                'https://drive.google.com/u/1/uc?export=download&confirm=8g0K&id=1cC5i6hplYq3j13plJc3x6BcvmEskC_bp', 'food_model.tar.gz')

        self.tokenizer = tfds.features.text.SubwordTextEncoder.load_from_file(
            './' + model_name +'/v1/token')

        nb_classes = 29 if model_name == 'category_model' else 2

        cnn = CNN(vocab_size=self.tokenizer.vocab_size, emb_dim=300, nb_filters=100,
            ffn_units=256, nb_classes=nb_classes, dropout_rate=0.2)

        cnn.load_weights('./' + model_name +'/v1/model_checkpoint/cp.ckpt').expect_partial()
        return cnn

    def inference_from_category_model(
        self, model, ingredients=[''], allergics=[''], description=[''], product_name=['']): 
        """
        This method is responsible to infer the product category based on your caracteristics. Pass just one product
        by time.

        Args:
            ingredients (list): ['água açúcar baunilha']
            allergics (list): ['contém glúten, pode conter nozes']
            description (list):  ['Esse produto é simplesmente fantástico']
            product_name (list):  ['energético sabor baunilha']

        Returns:
            str: Category name. e.g: 'Bebidas não alcoolicas'

        """     
        X = self.utils.get_cleaned_predictor([ingredients], [allergics], [description], [product_name])
        X = self.preprocessing_fn(X, './category_model/v1/token')

        label_encoder = self.get_label_encoder('./category_model/v1/label_encoder.pkl')

        product_category = model(np.array(X), training=False).numpy()[0]

        product_category_proba = product_category[product_category.argmax()]

        if product_category_proba < 0.6:
            product_category = 'categoria desconhecida'
        else:
            product_category = label_encoder.inverse_transform([product_category.argmax()])[0]

        return product_category

    def inference_from_food_model(self, model, product_name):
        """
        This method is responsible to infer if a product is a food or not. Pass just one product
        by time.

        Args:
            product_name (list):  ['energético sabor baunilha']

        Returns:
            str: food or not food.

        """     
        X = self.utils.get_description_and_name(product_name)
        X = self.preprocessing_fn([X], './food_model/v1/token')

        label_encoder = self.get_label_encoder('./food_model/v1/label_encoder.pkl')

        product_type = model(np.array(X), training=False).numpy()
        product_type = 1 if product_type > 0.5 else 0

        return label_encoder.inverse_transform([product_type])[0] 

    def download_model(self, url, output):
        gdown.download(url, output, quiet=False) 

        tar = tarfile.open(output)
        tar.extractall()
        tar.close()

    def preprocessing_fn(self, X, token_path):
        self.tokenizer = tfds.features.text.SubwordTextEncoder.load_from_file(token_path)
        tokenized_predictor = [self.tokenizer.encode(sentence) for sentence in  X] 

        return tokenized_predictor

    def get_label_encoder(self, encoder_path):
        pkl_encoder_file = open(encoder_path, 'rb')
        encoder = pickle.load(pkl_encoder_file) 
        pkl_encoder_file.close()
        return encoder

classifyProducts()



from datetime import datetime
import sys
import os 
from sys import argv

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import json
import pymongo
import pandas as pd

from database.db_mongo import MongoDatabase
from database.db_neo4j import Neo4jDatabase

class dataframeGen(object):

    def __init__(self, environment):
        self.environment = environment
        self.create_dataframe()

    
    def create_dataframe(self):
        self.open_database_connections()

        products_categories = self.get_products_categories()
        self.products_dataframe = self.get_products_dataframe(
            products_categories)

        self.save_dataframe(self.products_dataframe)

        self.close_mongo_connection()

    def open_database_connections(self):
        self.mongo = MongoDatabase(self.environment)
        self.mongo.connect()
        self.mongo_conn = self.mongo

        self.neo4j = Neo4jDatabase(self.environment)
        self.neo4j_conn = self.neo4j.connect()

    def get_products_categories(self):
        query = ''' MATCH (c:Category) return c '''
        products_categories = self.neo4j.match(query, self.neo4j_conn)

        return products_categories

    def get_products_dataframe(self, products_categories):
        df_products = pd.DataFrame()

        query = ('''MATCH (p:Product)-[:FROM_CATEGORY]-(c:Category {name: $name})
           OPTIONAL MATCH (i:Ingredient)<-[:HAS_INGREDIENT]-(p)
                   RETURN p as product, i as ingredient, c as category''')

        for category in products_categories:

            products_from_category = self.neo4j.match(
                query, self.neo4j_conn, {'name': category[0]['name']})

            products_and_ingredients = self.create_products_dict(
                products_from_category)

            category_dataframe = self.create_product_dataframe(
                products_and_ingredients)

            category_dataframe['Category'] = category[0]['name']

            df_products = pd.concat([df_products, category_dataframe], sort=True)

            print('Finished Category: ', category[0]['name'])
            sys.stdout.flush()
        return df_products

    def create_products_dict(self, products_and_ingredients):
        products_and_ingredients_dict = {}

        for record in products_and_ingredients:
            product_name = record['product']['name']

            product_ingredient = record.get(
                'ingredient') if record.get('ingredient') != None else {}

            product_ingredient = product_ingredient.get(
                'name') if product_ingredient != {} else ''

            if product_name in products_and_ingredients_dict:
                products_and_ingredients_dict[product_name].append(
                    product_ingredient)
            else:
                products_and_ingredients_dict[product_name] = [
                    record['product'], product_ingredient]

        return products_and_ingredients_dict

    def create_product_dataframe(self, data):
        df_data = []

        for product in data.values():
            products_data = {}
            products_data = product[0]._properties
            products_data['ingredients'] = product[1:len(product)]
            df_data.append(products_data)

        product_dataframe = pd.DataFrame(df_data)

        return product_dataframe

    def save_dataframe(self, categories_dataframe):

        payload = json.loads(
            categories_dataframe.to_json(orient='records'))

        self.mongo_conn.db['df_products'].insert_one({
            'df': payload,
            'created': pd.datetime.now()
        })

    def close_mongo_connection(self):
        self.mongo.close_connection()
        
dataframeGen(argv[1] if len(argv) > 1 else 'dev')


import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../'))

import tensorflow as tf
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix

import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from utils import utils

class trainProductsClassifier(object):

    def __init__(self, show_confusion_matrix=True):
        self.confusion_matrix = show_confusion_matrix
        self.train_model()

    def train_model(self):
        utils.create_saved_model_folder('food_model')

        dataframe, y = self.get_data()
        X = self.get_previsors_attrib(dataframe)
        
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.3,
                                                            stratify = y) 

        cnn = utils.create_model_instance(y, self.tokenizer)
        cnn.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])

        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath='./src/nkia/ml/saved_model/food_model/v1/model_checkpoint/cp.ckpt',
                                                        save_weights_only=True,
                                                        verbose=1)

        cnn.fit(X_train, y_train,
                batch_size = 64,
                epochs = 1,
                verbose = 1,
                validation_split = 0.10,
                callbacks=[cp_callback])

        utils.evaluate_model(cnn, X_test, y_test)

        if self.confusion_matrix:
            y_pred_test = cnn.predict(X_test)
            y_pred_test[y_pred_test > 0.5] = 1
            y_pred_test[y_pred_test <= 0.5] = 0
            utils.show_confusion_matrix(y_test, y_pred_test)

    def get_data(self):
        dataframe = pd.read_excel('./src/utils/datasets/positive_and_negative_food.xlsx')

        y = utils.encode_class_value(
            dataframe['Category'].values, './src/nkia/ml/saved_model/food_model/v1/label_encoder.pkl')

        print('Dataframe read')
        sys.stdout.flush()
        return dataframe, y

    def get_previsors_attrib(self, dataframe): 
        product_name = [[x] for x in dataframe['name']]

        X = list(map(utils.get_description_and_name, product_name))

        X, self.tokenizer = utils.preprocessing_fn(
            X, './src/nkia/ml/saved_model/food_model/v1/token')

        return X

trainProductsClassifier()



import os
from termcolor import colored

def mongo_credentials(environment):
    """
    Método responsável por retornar as credenciais do banco MONGODB de acordo com o ambiente de execução
    """
    credentials = {
        "uri": os.getenv("MONGO_URI_CONN") or "mongodb://localhost:27017/test",
        "ssl": os.getenv("MONGO_SSL") or True
    }

    if not credentials["uri"]:
        print(
            colored("Erro: ", "red"),
            colored(
                "MONGODB: Variável de ambiente uri não definida", "yellow"
            )
        )

    return credentials

def neo4j_credentials(environment):
    """
    Método responsável por retornar as credenciais do banco NEO4J de acordo com o ambiente de execução
    """
    credentials = {
        "host": os.getenv("BOT_NEO4J_HOST") or 'localhost',
        "port": os.getenv("BOT_NEO4J_PORT") or 7687,
        "user": os.getenv("BOT_NEO4J_USER") or 'neo4j',
        "db_name": os.getenv("BOT_NEO4J_NAME") or 'neo4j',
        "passwd": os.getenv("BOT_NEO4J_PASSWORD") or 'neo4j'

    }

    return credentials


import pandas as pd
import pymongo
import unidecode
import re 
import string
import nltk
import sys
import tensorflow_datasets as tfds
import tensorflow as tf
import pickle
import os
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import seaborn as sns
from nkia.ml.cnn_nlp_model import CNN

def get_dataframe_from_mongo(mongo_conn):
    products_df = mongo_conn.db.df_products.find({}).sort('created', pymongo.DESCENDING)
    
    if products_df:
        return pd.DataFrame(products_df[0]['df'])

    return ''

def get_cleaned_predictor(ingredients, allergics, description, product_name):
    ingredients_and_allergics = list(map(get_ingredients_and_allergics, ingredients))
    description = list(map(get_description_and_name, description))
    product_name = list(map(get_description_and_name, product_name))

    X = join_all_predictors(ingredients_and_allergics, description, product_name)

    return X

def get_ingredients_and_allergics(ingredients):
    # ingredients += allergics if allergics != '[]' else ''   
    ingredient = list(map(lambda x: unidecode.unidecode(x).lower(), ingredients))
    
    return ' '.join(ingredient)

def get_description_and_name(data):    
    stop_words = nltk.corpus.stopwords.words('portuguese')

    data = unidecode.unidecode(data[0]).lower()
    data = re.sub(r'['+string.punctuation+']', '', data)
    data = [word for word in data.split() if word not in stop_words]
    
    return ' '.join(data)   

def join_all_predictors(ingredients_and_allergics, description, product_name):
    count = 0
    X = []

    for name in product_name:
        X.append(name + ' ' + description[count] + ' ' + ingredients_and_allergics[count])
        count += 1

    return X

def preprocessing_fn(X, token_path):
    X, tokenizer = tokenize_data(X, token_path)
    X = padding_matrix(X)

    print('Preprocessing_fn finished')
    sys.stdout.flush()
    return X, tokenizer

def tokenize_data(data, token_path):
    tokenizer = tfds.features.text.SubwordTextEncoder.build_from_corpus(
        data, target_vocab_size=2**14)

    tokenizer.save_to_file(token_path)
    data = [tokenizer.encode(sentence) for sentence in data]

    return data, tokenizer

def padding_matrix(data):
    max_sentence_len = max([len(sentence) for sentence in data])
    data = tf.keras.preprocessing.sequence.pad_sequences(data,
                                                            value=0,
                                                            padding='post',
                                                            maxlen=max_sentence_len)

    return data

def encode_class_value(class_value, encoder_path):
    encoder = LabelEncoder()
    y = encoder.fit_transform(class_value)

    output_encoder = open(encoder_path, 'wb')
    pickle.dump(encoder, output_encoder)
    output_encoder.close()
    
    return y

def evaluate_model(cnn, X_test, y_test):
    results = cnn.evaluate(X_test, y_test, batch_size=64)
    print('Loss and Accuracy: ', results)

def show_confusion_matrix(y_test, predicted):
    fig, ax = plt.subplots(figsize=(10,10))
    cm = confusion_matrix(y_test, predicted)
    sns.heatmap(cm, annot=True)
    plt.show()

def create_model_instance(y, tokenizer):
    vocab_size = tokenizer.vocab_size
    nb_classes = len(set(y))

    cnn = CNN(vocab_size=vocab_size, emb_dim=300, nb_filters=100,
        ffn_units=256, nb_classes=nb_classes, dropout_rate=0.2)

    print('Created model instance')
    sys.stdout.flush()
    return cnn

def create_saved_model_folder(model_name):
    if 'saved_model' not in os.listdir('./src/nkia/ml/'):
        os.makedirs('./src/nkia/ml/saved_model/' + model_name + '/v1')
    elif model_name not in os.listdir('./src/nkia/ml/saved_model'):
        os.makedirs('./src/nkia/ml/saved_model/' + model_name + '/v1')
    


import sys
import os
from sys import argv
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../'))

import tensorflow as tf
from sklearn.model_selection import train_test_split

from utils import utils
from database.db_mongo import MongoDatabase

class trainProductsClassifier(object):

    def __init__(self, environment, show_confusion_matrix=True):
        self.environment = environment
        self.confusion_matrix = show_confusion_matrix
        self.train_model()

    def train_model(self):
        utils.create_saved_model_folder('category_model')
        dataframe, y = self.get_data()
        X = self.get_previsors_attrib(dataframe)
        
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.3,
                                                            stratify=y) 

        cnn = utils.create_model_instance(y, self.tokenizer)
        cnn.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath='./src/nkia/ml/saved_model/category_model/v1/model_checkpoint/cp.ckpt',
                                                        save_weights_only=True,
                                                        verbose=1)

        cnn.fit(X_train, y_train,
                batch_size = 64,
                epochs = 5,
                verbose = 1,
                validation_split = 0.10,
                callbacks=[cp_callback])

        utils.evaluate_model(cnn, X_test, y_test)

        if self.confusion_matrix:
            y_pred_test = cnn.predict(X_test)
            predicted = [result.argmax() for result in y_pred_test]
            utils.show_confusion_matrix(y_test, predicted)

    def get_data(self):
        self.open_mongo_connections()
        dataframe = utils.get_dataframe_from_mongo(self.mongo_conn)
        self.close_mongo_connection()

        dataframe['description'].fillna('', inplace=True)

        y = utils.encode_class_value(
            dataframe['Category'].values, './src/nkia/ml/saved_model/category_model/v1/label_encoder.pkl')

        print('Restored data from Mongo')
        sys.stdout.flush()
        return dataframe, y

    def open_mongo_connections(self):
        self.mongo = MongoDatabase(self.environment)
        self.mongo.connect()
        self.mongo_conn = self.mongo

    def close_mongo_connection(self):
        self.mongo.close_connection()

    def get_previsors_attrib(self, dataframe):
        ingredients = [x for x in dataframe['ingredients']] 
        # allergics = [x for x in dataframe['allergics']]
        description = [[x] for x in dataframe['description']]
        product_name = [[x] for x in dataframe['name']]

        X = utils.get_cleaned_predictor(ingredients, [], description, product_name)
        
        X, self.tokenizer = utils.preprocessing_fn(
            X, './src/nkia/ml/saved_model/category_model/v1/token')

        return X


trainProductsClassifier(argv[1] if len(argv) > 1 else 'dev')
